"""
Control and add subsystems to the running daemon hub
"""
import os
from collections.abc import Mapping
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Iterator
from typing import List
from typing import Tuple

import pop.hub


def add(
    hub: pop.hub.Hub,
    pypath: List[str] or str = None,
    subname: str = None,
    sub: pop.hub.Sub = None,
    static: List[str] or str = None,
    contracts_pypath: List[str] or str = None,
    contracts_static: List[str] or str = None,
    default_contracts: List[str] or str = None,
    virtual: bool = True,
    dyne_name: str = None,
    omit_start: Tuple[str] = ("_",),
    omit_end: Tuple[str] = (),
    omit_func: bool = False,
    omit_class: bool = True,
    omit_vars: bool = False,
    mod_basename: str = "pop.sub",
    stop_on_failures: bool = False,
    init: bool = True,
    load_all: bool = True,
    recursive_contracts_static: List[str] or str = None,
    default_recursive_contracts: List[str]
    or str = None,  # TODO: Not str, pretty sure -W. Werner, 2020-10-20
):
    """
    Add a new subsystem to the hub
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    :param sub: The sub to use as the root to add to
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    :param contracts_pypath: Load additional contract paths
    :param contracts_static: Load additional contract paths from a specific directory
    :param default_contracts: Specifies that a specific contract plugin will be applied as a default to all plugins
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param dyne_name: The dynamic name to use to look up paths to find plugins -- linked to conf.py
    :param omit_start: Allows you to pass in a tuple of characters that would omit the loading of any object
        I.E. Any function starting with an underscore will not be loaded onto a plugin
        (You should probably never change this)
    :param omit_end:Allows you to pass in a tuple of characters that would omit the loading of an object
        (You should probably never change this)
    :param omit_func: bool: Don't load any functions
    :param omit_class: bool: Don't load any classes
    :param omit_vars: bool: Don't load any vars
    :param mod_basename: str: Manipulate the location in sys.modules that the plugin will be loaded to.
        Allow plugins to be loaded into a separate namespace.
    :param stop_on_failures: If any module fails to load for any reason, stacktrace and do not continue loading this sub
    :param init: bool: determine whether or not we process __init__ functions
    :param load_all: Load all the plugins on the sub
    """
    if pypath:
        pypath = pop.hub.ex_path(pypath)
        subname = subname if subname else pypath[0].split(".")[-1]
    elif static:
        subname = subname if subname else os.path.basename(static)
    if dyne_name:
        subname = subname if subname else dyne_name
    root = sub or hub
    root._subs[subname] = pop.hub.Sub(
        hub,
        subname,
        root,
        pypath,
        static,
        contracts_pypath,
        contracts_static,
        default_contracts,
        virtual,
        dyne_name,
        omit_start,
        omit_end,
        omit_func,
        omit_class,
        omit_vars,
        mod_basename,
        stop_on_failures,
        init=init,
        sub_virtual=getattr(root, "_subvirt", True),
        recursive_contracts_static=recursive_contracts_static,
        default_recursive_contracts=default_recursive_contracts,
    )
    # init the sub (init.py:__init__) after it can be referenced on the hub!
    root._subs[subname]._sub_init()
    root._iter_subs = sorted(root._subs.keys())
    if load_all:
        root._subs[subname]._load_all()
    for alias in root._subs[subname]._alias:
        root._sub_alias[alias] = subname


def remove(hub: pop.hub.Hub, subname: str):
    """
    Remove a pop from the hub, run the shutdown if needed
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    """
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        if hasattr(sub, "init"):
            mod = getattr(sub, "init")
            if hasattr(mod, "shutdown"):
                mod.shutdown()
        hub._remove_subsystem(subname)


def load_all(hub: pop.hub.Hub, subname: str) -> bool:
    """
    Load all modules under a given pop
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    """
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        sub._load_all()
        return True
    else:
        return False


def get_dirs(hub: pop.hub.Hub, sub: pop.hub.Sub) -> List[str]:
    """
    Return a list of directories that contain the modules for this subname
    :param hub: The redistributed pop central hub
    :param sub: The pop object that contains the loaded module data
    """
    return sub._dirs


def iter_subs(
    hub: pop.hub.Hub, sub: pop.hub.Sub, recurse: bool = False
) -> Generator[pop.hub.Sub, None, None]:
    """
    Return an iterator that will traverse just the subs. This is useful for
    nested subs
    :param hub: The redistributed pop central hub
    :param recurse: Recursively iterate over nested subs
    """
    for name in sorted(sub._subs):
        ret = sub._subs[name]
        if ret._sub_virtual:
            yield ret
            if recurse:
                if hasattr(ret, "_subs"):
                    yield from hub.pop.sub.iter_subs(ret, recurse)


def load_subdirs(hub: pop.hub.Hub, sub: pop.hub.Sub, recurse: bool = False):
    """
    Given a sub, load all subdirectories found under the sub into a lower namespace
    :param hub: The redistributed pop central hub
    :param sub: The pop object that contains the loaded module data
    :param recurse: Recursively iterate over nested subs
    """
    if not sub._sub_virtual:
        return
    dirs = hub.pop.sub.get_dirs(sub)
    roots = {}
    for dir_ in dirs:
        for fn in os.listdir(dir_):
            if fn.startswith("_"):
                continue
            if fn == "contracts":
                continue
            full = os.path.join(dir_, fn)
            if not os.path.isdir(full):
                continue
            if fn not in roots:
                roots[fn] = [full]
            else:
                roots[fn].append(full)
    for name, sub_dirs in roots.items():
        # Load er up!
        hub.pop.sub.add(
            subname=name,
            sub=sub,
            static=sub_dirs,
            virtual=sub._virtual,
            omit_start=sub._omit_start,
            omit_end=sub._omit_end,
            omit_func=sub._omit_func,
            omit_class=sub._omit_class,
            omit_vars=sub._omit_vars,
            mod_basename=sub._mod_basename,
            stop_on_failures=sub._stop_on_failures,
        )
        if recurse:
            if isinstance(getattr(sub, name), pop.hub.Sub):
                hub.pop.sub.load_subdirs(getattr(sub, name), recurse)


def reload(hub: pop.hub.Hub, subname: str):
    """
    Instruct the hub to reload the modules for the given sub. This does not call
    the init.new function or remove sub level variables. But it does re-read the
    directory list and re-initialize the loader causing all modules to be re-evaluated
    when started.
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    """
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        sub._prepare()
        return True
    else:
        return False


def extend(
    hub: pop.hub.Hub,
    subname: str,
    pypath: List[str] or str = None,
    static: List[str] or str = None,
    contracts_pypath: List[str] or str = None,
    contracts_static: List[str] or str = None,
) -> bool:
    """
    Extend the directory lookup for a given sub. Any of the directory lookup
    arguments can be passed.
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    :param contracts_pypath: Load additional contract paths
    :param contracts_static: Load additional contract paths from a specific directory
    """
    if not hasattr(hub, subname):
        return False
    sub = getattr(hub, subname)
    if pypath:
        sub._pypath.extend(pop.hub.ex_path(pypath))
    if static:
        sub._static.extend(pop.hub.ex_path(static))
    if contracts_pypath:
        sub._contracts_pypath.extend(pop.hub.ex_path(contracts_pypath))
    if contracts_static:
        sub._contracts_static.extend(pop.hub.ex_path(contracts_static))
    sub._prepare()
    return True


def dynamic(hub: pop.hub.Hub, sub: pop.hub.Sub, get_reference: Callable):
    """
    Create dynamic subs underneath the named sub.
    When a dynamic sub is called, determine the actual path that should be used
    and call the reference with the parameters passed to the Dynamic Sub

    :param hub: The redistributed pop central hub
    :param sub: The sub to modify to have dynamic subs underneath it
    :param get_reference: A callable function that returns the desired target base sub
    """

    class Dynamic(Mapping):
        def __init__(
            self,
            parent_sub: pop.hub.Sub,
            subs: Dict[str, pop.hub.Sub],
            ref: str = "",
        ):
            """
            :param parent_sub: The sub that owns the Dynamic subs
            :param subs: The subs that belonged to the parent sub
            :param ref: The names of Dynamic subs that led up to this reference
            """
            self._sub = parent_sub
            self._subs = subs
            self._ref = ref

        def __getitem__(self, name: str) -> pop.hub.Sub:
            """
            If the sub exists in the parent sub dictionary, return it directly.
            If the sub does not exist in the parent, then create a CallableSub
            that can be used to call the desired function when the ref has been fully built.
            :param name: The name of the sub to return
            """
            # Return the sub if it exists, this is a static reference
            if name in self._subs:
                return self._subs[name]
            else:
                # Check if one of our static subs has the name we are looking for
                for static_sub in self._subs.values():
                    if name in static_sub._alias:
                        return static_sub

            # We are working with an unresolved reference

            # Keep track of the dynamic reference chain that has led up to the current dynamic sub
            if self._ref:
                ref = f"{self._ref}.{name}"
            else:
                ref = name

            # In case the unresolved reference happens to be a function, create a callable sub
            callable_sub = CallableSub(
                hub=hub, subname=name, root=self._sub, init=False, ref=ref
            )

            # Any subsequent references are just as unresolved as this one, replace subs with Dynamic subs
            callable_sub._subs = Dynamic(callable_sub, {}, ref)
            return callable_sub

        # These methods maintain the functionality of being a mapping of sub-names to subs
        def __setitem__(self, name: str, new_sub: pop.hub.Sub):
            self._subs[name] = new_sub

        def __len__(self) -> int:
            return len(self._subs)

        def __iter__(self) -> Iterator[str]:
            return iter(self._subs)

    class CallableSub(pop.hub.Sub):
        """
        A Sub that transforms a Dynamic reference on the hub to a static reference and calls it.
        """

        def __init__(self, ref: str, *args, **kwargs):
            self._ref = ref
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            # Get the actual base reference based on user logic
            func = get_reference(*args, **kwargs)
            # Transform the built reference determined by parent dynamic subs into a real function located on the hub
            for ref in self._ref.split("."):
                func = getattr(func, ref)

            # Call the target function with the parameters passed to the callable sub
            return func(*args, **kwargs)

    # replace the _subs dict with a dynamically referenced version
    sub._subs = Dynamic(sub, sub._subs)



"""
Seed a new project with a directory tree and first files
"""
import os

import pop.hub

SETUP = """#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
from setuptools import setup, Command

NAME = "%%NAME%%"
DESC = ""

# Version info -- read without importing
_locals = {}
with open("{}/version.py".format(NAME)) as fp:
    exec(fp.read(), None, _locals)
VERSION = _locals["version"]
SETUP_DIRNAME = os.path.dirname(__file__)
if not SETUP_DIRNAME:
    SETUP_DIRNAME = os.getcwd()

with open("README.rst", encoding="utf-8") as f:
    LONG_DESC = f.read()

with open("requirements.txt") as f:
    REQUIREMENTS = f.read().splitlines()


class Clean(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for subdir in (NAME, "tests"):
            for root, dirs, files in os.walk(
                os.path.join(os.path.dirname(__file__), subdir)
            ):
                for dir_ in dirs:
                    if dir_ == "__pycache__":
                        shutil.rmtree(os.path.join(root, dir_))


def discover_packages():
    modules = []
    for package in (NAME,):
        for root, _, files in os.walk(os.path.join(SETUP_DIRNAME, package)):
            pdir = os.path.relpath(root, SETUP_DIRNAME)
            modname = pdir.replace(os.sep, ".")
            modules.append(modname)
    return modules


setup(
    name=NAME,
    author="",
    author_email="",
    url="",
    version=VERSION,
    install_requires=REQUIREMENTS,
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/x-rst",
    python_requires=">=3.6",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 5 - Production/Stable",
    ],
    packages=discover_packages(),
    entry_points={"console_scripts": ["%%NAME%% = %%NAME%%.scripts:start",],},
    cmdclass={"clean": Clean},
)
"""

PYPROJ = r"""[tool.black]
line-length = 88
target-version = ['py36', 'py37', 'py38']
include = '\.pyi?$'
exclude = '''
(
  /(
      \.eggs
    | \.git
    | \.hg
    | \.mypy_cache
    | \.tox
    | \.venv
    | _build
    | buck-out
    | build
    | dist
  )/
)
'''
"""

PRECOM = r"""---
minimum_pre_commit_version: 1.15.2
repos:
  - repo: https://github.com/pre-commit/pre-commit-hooks
    rev: v2.5.0
    hooks:
      - id: check-merge-conflict  # Check for files that contain merge conflict strings.
        language_version: python3
      - id: trailing-whitespace   # Trims trailing whitespace.
        args: [--markdown-linebreak-ext=md]
        language_version: python3
      - id: mixed-line-ending     # Replaces or checks mixed line ending.
        args: [--fix=lf]
        language_version: python3
      - id: end-of-file-fixer     # Makes sure files end in a newline and only a newline.
        exclude: tests/fake_.*\.key
        language_version: python3
      - id: check-ast             # Simply check whether files parse as valid python.
        language_version: python3
      - id: check-yaml
      - id: check-json
  -   repo: https://github.com/psf/black
      rev: 19.10b0
      hooks:
      - id: black
        language_version: python3
"""

ENTRY = """entry_points={
        'console_scripts': [
            '%%NAME%% = %%NAME%%.scripts:start',
            ],
          },"""

SCRIPT = """#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pop.hub


def start():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="%%NAME%%")
    hub.%%NAME%%.init.cli()
"""

INIT = """def __init__(hub):
    # Remember not to start your app in the __init__ function
    # This function should just be used to set up the plugin subsystem
    # Add another function to call from your run.py to start the app
    pass


def cli(hub):
    hub.pop.config.load(["%%NAME%%"], cli="%%NAME%%")
    print("%%NAME%% works!")
"""

REQ = "pop\n"

CONF = """CLI_CONFIG = {}
CONFIG = {}
SUBCOMMANDS = {}
DYNE = {
    "%%NAME%%": ["%%NAME%%"],
%%DYNE%%}
"""

VER = """# -*- coding: utf-8 -*-
version = "1"\n
"""


def new(hub: "pop.hub.Hub"):
    """
    Given the option in hub.opts "seed_name" create a directory tree for a
    new pop project
    """
    hub.PATH = os.getcwd()
    name = hub.opts["seed_name"]
    for dyne in hub.opts["dyne"]:
        hub.pop.seed.mkdir(name, dyne)
        hub.pop.seed.mkdir(name, dyne, "contracts")
    if hub.opts["type"] == "v":
        hub.pop.seed.mkdir(name)
        hub.pop.seed.mksetup(name, entry=False)
        hub.pop.seed.mkversion(name)
        hub.pop.seed.mkconf(name)
        hub.pop.seed.mkreq(name)
        hub.pop.seed.mkreadme(name)
    else:
        hub.pop.seed.mkdir(name, name)
        hub.pop.seed.mkdir(name, name, "contracts")
        hub.pop.seed.mksetup(name)
        hub.pop.seed.mkscript(name)
        hub.pop.seed.mkversion(name)
        hub.pop.seed.mkconf(name)
        hub.pop.seed.mkreq(name)
        hub.pop.seed.mkrun(name)
        hub.pop.seed.mkinit(name)
        hub.pop.seed.mkreadme(name)
    hub.pop.seed.mkproj()
    hub.pop.seed.mkprecom()
    hub.pop.seed.print_post(name)


def mkdir(hub: "pop.hub.Hub", *args):
    """
    Create the named dir
    """
    path = hub.PATH
    for dir_ in args:
        path = os.path.join(path, dir_)
        if not os.path.isdir(path):
            try:
                os.makedirs(path)
            except Exception:
                print(f"Failed to make {path}")
                continue
            if dir_ == "scripts" and len(args) == 1:
                continue


def mkreq(hub: "pop.hub.Hub", name: str):
    path = os.path.join(hub.PATH, "requirements.txt")
    with open(path, "w+") as fp:
        fp.write(REQ)


def mksetup(hub: "pop.hub.Hub", name: str, entry: bool = True):
    """
    Create and write out a setup.py file
    """
    path = os.path.join(hub.PATH, "setup.py")
    setup_str = SETUP.replace("%%NAME%%", name)
    if entry:
        setup_str = setup_str.replace(
            "%%ENTRY%%",
            ENTRY.replace(
                "%%NAME%%.scripts:start", f"{name.replace('-', '_')}.scripts:start"
            ),
        )
        setup_str = setup_str.replace("%%ENTRY%%", ENTRY.replace("%%NAME%%", name))
    else:
        setup_str = setup_str.replace("%%ENTRY%%", "")
    with open(path, "w+") as fp:
        fp.write(setup_str)


def mkscript(hub: "pop.hub.Hub", name: str):
    """
    Create and write out a setup.py file
    """
    path = os.path.join(hub.PATH, name, "scripts.py")
    script_str = SCRIPT.replace("%%NAME%%", name)
    with open(path, "w+") as fp:
        fp.write(script_str)


def mkrun(hub: "pop.hub.Hub", name: str):
    """
    Create the convenience run.py script allowing the project to
    be executed from the local directory
    """
    path = os.path.join(hub.PATH, "run.py")
    run_str = SCRIPT.replace("%%NAME%%", name)
    run_str += "\n\nstart()\n"
    with open(path, "w+") as fp:
        fp.write(run_str)


def mkinit(hub: "pop.hub.Hub", name: str):
    """
    Create the intial init.py
    """
    path = os.path.join(hub.PATH, name, name, "init.py")
    init_str = INIT.replace("%%NAME%%", name)
    with open(path, "w+") as fp:
        fp.write(init_str)


def mkversion(hub: "pop.hub.Hub", name: str):
    """
    Create the version.py file
    """
    path = os.path.join(hub.PATH, name, "version.py")
    with open(path, "w+") as fp:
        fp.write(VER)


def mkconf(hub: "pop.hub.Hub", name: str):
    """
    Create the version.py file
    """
    path = os.path.join(hub.PATH, name, "conf.py")
    dyne_str = ""
    for dyne in hub.opts["dyne"]:
        dyne_str += f'    "{dyne}": ["{dyne}"],\n'
    conf_str = CONF.replace("%%NAME%%", name)
    conf_str = conf_str.replace("%%DYNE%%", dyne_str)
    with open(path, "w+") as fp:
        fp.write(conf_str)


def mkreadme(hub: "pop.hub.Hub", name: str):
    """
    Create and write out a setup.py file
    """
    path = os.path.join(hub.PATH, "README.rst")
    eqchars = "=" * len(name)
    readme_str = f"{eqchars}\n{name.upper()}\n{eqchars}\n"
    with open(path, "w+") as fp:
        fp.write(readme_str)


def mkproj(hub: "pop.hub.Hub"):
    """
    Create the pyproject.toml file
    """
    path = os.path.join(hub.PATH, "pyproject.toml")
    with open(path, "a+") as fp:
        fp.write(PYPROJ)


def mkprecom(hub: "pop.hub.Hub"):
    """
    Create the precommit file
    """
    path = os.path.join(hub.PATH, ".pre-commit-config.yaml")
    with open(path, "w+") as fp:
        fp.write(PRECOM)


def print_post(hub: "pop.hub.Hub", name: str):
    """
    Print a message after the run to document how to enable
    things like pre-commit
    """
    print(f"Congratulations! You now have a project set up called {name}!")
    print("This project can be executed by calling the run.py script:")
    print("    python3 run.py")
    print(
        "This project has been set up with pre-commit hooks for code checks and black."
    )
    print('First set up your source control environment with "git init" or "hg init".')
    print("Then enable these checks in your git checkout:")
    print("    pip install pre-commit")
    print("    pre-commit install")
    print("To run pre-commit manually, execute:")
    print("    pre-commit run --all-files")



"""
Used to take care of the options that end in `_dir`. The assumption is that
`_dir` options need to be treated differently. They need to verified to exist
and they need to be rooted based on the user, root option etc.
"""
import os

import pop.hub


def roots(hub: "pop.hub.Hub", default_root, f_opts, root_dir):
    """
    Detect the root dir data and apply it
    """
    os_root = os.path.abspath(os.sep)
    root = os_root
    change = False
    non_priv = False
    if hasattr(os, "geteuid"):
        if not os.geteuid() == 0:
            change = True
            non_priv = True
    if root_dir and root_dir != default_root:
        root = root_dir
        change = True
    if not root.endswith(os.sep):
        root = f"{root}{os.sep}"
    if change:
        for imp in f_opts:
            for key in f_opts[imp]:
                if key == "root_dir":
                    continue
                if key.endswith("_dir"):
                    if non_priv:
                        root = os.path.join(os.environ["HOME"], f".{imp}{os.sep}")
                    if imp in f_opts[imp][key]:
                        a_len = len(imp) + 1
                        f_opts[imp][
                            key
                        ] = f"{os_root}{f_opts[imp][key][f_opts[imp][key].index(imp)+a_len:]}"
                    f_opts[imp][key] = f_opts[imp][key].replace(os_root, root, 1)


def verify(hub: "pop.hub.Hub", opts):
    """
    Verify that the environment and all named directories in the
    configuration exist
    """
    for key in opts:
        if key == "root_dir":
            continue
        if key == "config_dir":
            continue
        if key.endswith("_dir"):
            if not os.path.isdir(opts[key]):
                os.makedirs(opts[key])



#!/usr/bin/python3
import pop.hub


def pop_seed():
    CONFIG = {
        "seed_name": {
            "positional": True,
            "help": "The name of the project that is being created",
        },
        "type": {
            "default": "p",
            "options": ["-t"],
            "help": 'The type of project to build, by default make a standalone project, but for a vetical app project pass a "v"',
        },
        "dyne": {
            "options": ["-d"],
            "default": [],
            "nargs": "*",
            "help": "A space delimited list of additional dynamic names for vertical app-merging",
        },
    }

    hub = pop.hub.Hub()
    hub.pop.sub.add("pop.mods.conf")
    hub.opts = hub.conf.reader.read(CONFIG)
    hub.pop.seed.new()



"""
Convenience wrappers to make using the conf system as easy and seamless as possible
"""
from typing import Any
from typing import Dict
from typing import List

import pop.hub


def integrate(
    hub: "pop.hub.Hub",
    imports: List[str] or str,
    override: Dict[str, Any] = None,
    cli: str = None,
    roots: bool = None,
    loader: str = "json",
    logs: bool = True,
):
    """
    Load the conf sub and run the integrate sequence.
    """
    hub.pop.sub.add("pop.mods.conf")
    hub.conf.integrate.load(
        imports, override, cli=cli, roots=roots, loader=loader, logs=logs
    )



"""
Tools to work with dicts
"""
import warnings

import dict_tools.data
import dict_tools.update

__virtualname__ = "dicts"


def update(*args, **kwargs):
    warnings.warn(
        "Use `dict_tools.update.update()` instead", DeprecationWarning, stacklevel=1
    )
    return dict_tools.update.update(*args, **kwargs)


def traverse(*args, **kwargs):
    warnings.warn(
        "Use `dict_tools.data.traverse_dict_and_list()` instead", DeprecationWarning
    )
    return dict_tools.data.traverse_dict_and_list(*args, **kwargs)



"""
Integrate is used to pull config data from multiple sources and merge it into
the hub. Once it is merged then when a sub is loaded the respective config data
is loaded into the sub as `OPTS`
"""
# Take an *args list of modules to import and look for conf.py
# Import conf.py if present
# After gathering all dicts, modify them to merge CLI options
#
import copy
import importlib
import os


def _ex_final(confs, final, override, key_to_ref, ops_to_ref):
    """
    Scan the configuration datasets, create the final config
    value, and detect collisions
    """
    for arg in confs:
        for key in confs[arg]:
            ref = f"{arg}.{key}"
            if ref in override:
                s_key = override[ref]["key"]
                s_opts = override[ref]["options"]
            else:
                s_key = key
                s_opts = confs[arg][key].get("options", [])
            s_opts.append(f"--{s_key}")
            final[s_key] = confs[arg][key]
            if s_opts:
                final[s_key]["options"] = s_opts
            if s_key in key_to_ref:
                key_to_ref[s_key].append(ref)
            else:
                key_to_ref[s_key] = [ref]
            for opt in s_opts:
                if opt in ops_to_ref:
                    ops_to_ref[opt].append(ref)
                else:
                    ops_to_ref = [ref]


def load(
    hub,
    imports,
    override=None,
    cli=None,
    roots=False,
    loader="json",
    logs=True,
    version=True,
):
    """
    This function takes a list of python packages to load and look for
    respective configs. The configs are then loaded in a non-collision
    way modifying the cli options dynamically.
    The args look for the named <package>.conf python module and then
    looks for dictionaries named after the following convention:

    override = {'<package>.key': 'key': 'new_key', 'options': ['--option1', '--option2']}

    CONFIG: The main configuration for this package - loads to hub.OPT['<import>']
    CLI_CONFIG: Loaded only if this is the only import or if specified in the cli option
    SUBS: Used to define the subcommands, only loaded if this is the cli config
    """
    if override is None:
        override = {}
    if isinstance(imports, str):
        if cli is None:
            cli = imports
        imports = [imports]
    primary = imports[0] if cli is None else cli
    confs = {}
    final = {}
    collides = []
    key_to_ref = {}
    ops_to_ref = {}
    subs = {}
    for imp in imports:
        try:
            cmod = importlib.import_module(f"{imp}.conf")
        except ImportError:
            continue
        if hasattr(cmod, "CONFIG"):
            confs[imp] = copy.deepcopy(cmod.CONFIG)
        if cli == imp:
            if hasattr(cmod, "CLI_CONFIG"):
                confs[imp].update(copy.deepcopy(cmod.CLI_CONFIG))
            if hasattr(cmod, "SUBS"):
                subs = copy.deepcopy(cmod.SUBS)
    if logs:
        lconf = hub.conf.log.init.conf(primary)
        lconf.update(confs[primary])
        confs[primary] = lconf
    if version:
        vconf = hub.conf.version.CONFIG
        vconf.update(confs[primary])
        confs[primary] = vconf
    _ex_final(confs, final, override, key_to_ref, ops_to_ref)
    for opt in ops_to_ref:
        g_count = 0
        if len(ops_to_ref[opt]) > 1:
            collides.append({opt: ops_to_ref[opt]})
    for key in key_to_ref:
        col = []
        for ref in key_to_ref[key]:
            col.append(ref)
        if len(col) > 1:
            collides.append({key: key_to_ref[key]})
    if collides:
        raise KeyError(collides)
    opts = hub.conf.reader.read(final, subs, loader=loader)
    # This will be put into an immutable data type before it is passed on
    f_opts = {}
    for key in opts:
        if key == "_subparser_":
            f_opts["_subparser_"] = opts["_subparser_"]
            continue
        for ref in key_to_ref[key]:
            imp = ref[: ref.rindex(".")]
            local_key = ref[ref.rindex(".") + 1 :]
            if imp not in f_opts:
                f_opts[imp] = {}
            f_opts[imp][local_key] = opts[key]
    if roots:
        root_dir = f_opts.get(cli, {}).get("root_dir")
        hub.conf.dirs.roots(
            final.get("root_dir", {}).get("default", os.path.abspath(os.sep)),
            f_opts,
            root_dir,
        )
        for imp in f_opts:
            hub.conf.dirs.verify(f_opts[imp])
    hub.OPT = hub.pop.data.imap(f_opts)
    if logs:
        log_plugin = hub.OPT[primary].get("log_plugin")
        getattr(hub, f"conf.log.{log_plugin}.setup")(hub.OPT[primary])
    if hub.OPT[primary].get("version"):
        hub.conf.version.run(primary)



"""
Provides tools to help unit test projects using pop.
For now, provides mock Hub instances.
"""
import copy
import inspect
from asyncio import iscoroutinefunction

try:
    HAS_TEST = False
    from asynctest.mock import create_autospec, Mock

    HAS_TEST = True
except (ImportError, ModuleNotFoundError):
    try:
        from unittest.mock import create_autospec as mock_create_autospec, Mock

        HAS_TEST = True

        def create_autospec(spec, *args, **kwargs):
            if iscoroutinefunction(spec):
                raise Exception(
                    "MockHub requires asynctest in order to mock async functions"
                )
            return mock_create_autospec(spec, *args, **kwargs)

    except (ImportError, ModuleNotFoundError):
        ...

from pop.contract import Contracted
from pop.loader import LoadedMod
from pop.hub import Hub, Sub
from typing import Any, Callable, Tuple


def __virtual__(hub: Hub) -> Tuple[bool, str]:
    return HAS_TEST, "Async pop testing libs are not available"


class _LookUpTable:
    def __init__(self, *args, **kwargs):
        self._lut = {}
        super().__init__(*args, **kwargs)

    def __contains__(self, key: str):
        return id(key) in self._lut

    def __setitem__(self, key: str, value: Any):
        self._lut[id(key)] = value

    def __getitem__(self, key: str):
        return self._lut[id(key)]

    def __delitem__(self, key: str):
        del self._lut[id(key)]

    def __len__(self):
        return len(self._lut)


class _LazyPop:
    __lazy_classes = [Hub, Sub, LoadedMod]
    _hub_id = object()  # just a unique object for our lut storage
    _lazy_hub_id = object()

    class __Lazy:
        pass

    def __init__(self, obj, lut=None):
        if isinstance(obj, Hub):
            lut = _LookUpTable()
            lut[self._hub_id] = obj
            lut[self._lazy_hub_id] = self
            lut[obj] = self
        elif isinstance(obj, Sub):
            obj._load_all()

        self.__lut = lut
        self.__obj = obj
        for attr_name in self.__attr_names():
            setattr(self, attr_name, _LazyPop.__Lazy)

    def _hub(self):
        return self.__lut[self._hub_id]

    def _lazy_hub(self):
        return self.__lut[self._lazy_hub_id]

    def __attr_names(self):
        # TODO: '_' - is this actually right? what should I really expose?
        attrs = [attr for attr in self.__obj.__dict__ if not attr.startswith("_")]

        if isinstance(self.__obj, Hub):
            attrs += list(self.__obj._subs)
        elif isinstance(self.__obj, Sub):
            attrs += list(self.__obj._loaded)
            attrs += list(self.__obj._subs)
        elif isinstance(self.__obj, LoadedMod):
            attrs += list(self.__obj._attrs)
        else:
            raise Exception(
                "Standard objects should not be lazy: {}".format(str(self.__obj))
            )

        return attrs

    def _find_subs(self):
        i = 0
        subs = [(s._subname, s) for s in self._hub()]
        while i < len(subs):
            for child in subs[i][1]._subs:
                subs.append((".".join([subs[i][0], child]), getattr(subs[i][1], child)))
            i += 1
        return subs

    def _find_module_from_file(self, file):
        for path, sub in self._find_subs():
            try:
                mod = sub._vmap[file]
                return ".".join([path, mod]), getattr(sub, mod)
            except (AttributeError, KeyError):
                pass
        else:
            raise Exception("Module not loaded on hub.")

    def _reset(self):
        # A potential issue - we don't do reference counting, so it's *possible*
        #  that an object we're clearing out has been accessed via two places on
        #  the test hub:
        #
        #     mock_hub.OBJ.return_value = True
        #     mock_hub.pop.OBJ
        #
        #  If VAL is the same object on the real hub, and is accessed from *both*
        #   places on a test hub, mock_hub.pop._reset() will break the coupling -
        #   mock_hub.VAL and mock_hub.pop.VAL will then refer to different objects.
        #  Fixing this requires storing back-references in the LUT. (TODO)
        #  Problems arising this should be rare and also fairly obvious.

        # first, reset items with entries in the LUT
        items = list(self.__dict__.items())
        for k, v in items:
            orig = getattr(self.__obj, k, None)
            if orig in self.__lut and orig is not self.__obj:
                if isinstance(v, _LazyPop):
                    v._reset()
                del self.__lut[orig]
                setattr(self, k, _LazyPop.__Lazy)

        # now remove any assignments that *wouldn't* be in the lut
        # (attrs that weren't assigned on the backing object)
        fresh_obj = self.__class__(self.__obj, lut={})
        fresh_obj_keys = list(self.__dict__.keys())
        for k in fresh_obj_keys:
            if k not in fresh_obj.__dict__:
                del self.__dict__[k]

    def __setattr__(self, key, value):
        if isinstance(value, _LazyPop) and value.__class__ is not self.__class__:
            if value.__obj in self.__lut:
                # we've previously touched this item, we need to clear it out.
                self.__lut[value.__obj]._reset()
            # we are constructing a hybrid _LazyPop - copy the type of the value but nothing else
            value = value.__class__(value.__obj, self.__lut)
            self.__lut[value.__obj] = value
        if isinstance(value, Contracted) and value.hub is not self._lazy_hub():
            # We need to update the Contracted to use our parent hub
            value = copy.copy(value)
            value.hub = self._lazy_hub()
            # we don't update the lut because we don't know what the *original is *
        if value.__class__ in self.__lazy_classes and "__obj" not in key:
            raise TypeError(
                "Mixing of real and test hubs is not supported. "
                "Contracteds (hub.sub.mod.func) are supported."
            )

        super().__setattr__(key, value)

    def __getattribute__(self, item):
        if item and not item.strip("_"):  # only contains underscores, resolve 'this'
            stack = inspect.stack(0)
            file = stack[1].filename
            path, mod = self._find_module_from_file(file)

            # go up N steps
            parts = path.split(".")
            resolved_path = parts[0 : len(parts) - len(item) + 1]
            if resolved_path:
                orig = getattr(self._hub(), ".".join(resolved_path))
            else:
                orig = self._hub()

            # find/create attr, return
            attr = self._orig_to_attr(orig)
            self.__lut[orig] = attr
            return attr

        if "." in item:
            result = self
            for part in item.split(".").copy():
                result = getattr(result, part)
            return result

        attr = super().__getattribute__(item)

        if attr is _LazyPop.__Lazy:
            orig = getattr(self.__obj, item)
            attr = self._orig_to_attr(orig)

            self.__lut[orig] = attr
            super().__setattr__(item, attr)  # bypass our custom setattr

        return attr

    def __getitem__(self, item):
        return getattr(self, item)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def _orig_to_attr(self, orig):
        if orig in self.__lut:
            attr = self.__lut[orig]
        elif [True for cls in self.__lazy_classes if isinstance(orig, cls)]:
            attr = self.__class__(orig, self.__lut)
        elif isinstance(orig, Contracted):
            attr = self._mock_function(orig)
        else:
            attr = self._mock_attr(orig)
        return attr

    def _mock_attr(self, a):
        return create_autospec(a, spec_set=True)

    def _mock_function(self, f):
        raise NotImplementedError()


def strip_hub(f: Callable) -> Hub:
    """
    returns a no-op function with the same function signature... minus the first parameter (hub).
    """
    if inspect.iscoroutinefunction(f):
        newf = "async "
    else:
        newf = ""
    newf += f"def {f.__name__}("
    params = inspect.signature(f).parameters
    new_params = []
    for param in params:
        if params[param].kind is inspect.Parameter.VAR_POSITIONAL:
            new_params.append(f"*{param}")
        elif params[param].kind is inspect.Parameter.VAR_KEYWORD:
            new_params.append(f"**{param}")
        else:
            new_params.append(param)
        if params[param].default is not inspect.Parameter.empty:
            new_params[-1] += '="has default"'
    newf += ", ".join(new_params[1:])  # skip hub
    newf += "): pass"

    scope = {}
    exec(newf, scope)

    return scope[f.__name__]


def mock_hub(hub: Hub) -> "MockHub":
    return MockHub(hub)


class MockHub(_LazyPop):
    """
    Provides mocks mirroring a real hub::

        hub.sub.mod.fn()  # mock
        hub.sub.mod.attr  # mock
    """

    def _mock_function(self, f: Contracted) -> Callable:
        afunc = create_autospec(strip_hub(f.func), spec_set=True)
        afunc.__signature__ = f.signature
        return afunc


def fn_hub(hub: Hub) -> "NoContractHub":
    return NoContractHub(hub)


class NoContractHub(_LazyPop):
    """
    Provides access to real functions, bypassing contracts and mocking attributes::

        hub.sub.mod.fn()  # executes real function, no contracts
        hub.sub.mod.attr  # mock
    """

    def _mock_function(self, f: Contracted) -> Contracted:
        return Contracted(
            hub=self._lazy_hub(),
            contracts=None,
            func=f.func,
            ref=f.ref,
            name=f.__name__,
        )


def mock_contracted(contract_hub, c: Contracted) -> Contracted:
    mock_func = create_autospec(c.func, spec_set=True)
    mock_func.__signature__ = c.signature  # required for python 3.6
    mock_func.__module__ = c.func.__module__
    mock_func.__dict__.update(copy.deepcopy(c.func.__dict__))
    return Contracted(contract_hub, c.contracts, mock_func, c.ref, c.__name__)


class ContractHub(_LazyPop):
    """
    Runs a call through the contract system, but the function is a mock. Mostly useful for integration tests:

        hub.sub.mod.fn()  # executes mock function, real contracts
        hub.sub.mod.attr  # mock

    You can verify what parameters are passed to a function after going through loaded contracts::

        contract_hub.sub.mod.fn('foo')
        assert contract_hub.sub.mod.fn.called_with('bar')

    --------------------------------

    You can view or modify the contracts that will be executed on one function for a test - but first:
    MODIFYING CONTRACTS THIS WAY IS NOT SAFE ON REAL HUBS AND OTHER TESTING HUB VARIANTS!

    I have previously thought of modifying contracts with mocks, only to realize what I really want is to
    unit test a specific contract. Think twice before using this functionality.

    --------------------------------

    The contract modules are visible via hub.sub.mod.fn.contracts, and the contract functions that will
    be called, wrapping fn are visible via hub.sub.mod.fn.contract_functions. It is safe to modify the
    contracts list or contract_functions dict only on a ContractHub.

    Examine that the first contract function to be called is 'foo.pre_fn', then bypass it::

        assert contract_hub.sub.mod.fn.contract_functions['pre'][0].__module__ is 'foo'
        assert contract_hub.sub.mod.fn.contract_functions['pre'][0].__name__ is 'pre_fn'
        hub.sub.mod.fn.contract_functions['pre'][0] = create_autospec(hub.sub.mod.fn.contract_functions['pre'][0])

    Assert that one contract will be called before another::

        assert contract_hub.sub.mod.fn.contracts.index(contract1) < contract_hub.sub.mod.fn.contracts.index(contract2)
    """

    def _mock_function(self, f: Contracted) -> Contracted:
        return mock_contracted(self._lazy_hub(), f)


def mock_attr_hub(hub: Hub) -> "MockAttrHub":
    return MockAttrHub(hub)


class MockAttrHub(_LazyPop):
    """
    Provides an almost-normal hub.

    Contracts and functions are executed, but all attrs are mocked.
    """

    def _mock_function(self, f) -> Contracted:
        return Contracted(
            hub=self._lazy_hub(),
            contracts=f.contracts,
            func=f.func,
            ref=f.ref,
            name=f.__name__,
        )



"""
Translate an options data structure into command line args
"""
import argparse
import collections
import functools
import inspect
import sys

import pop.hub

__virtualname__ = "args"
__contracts__ = [__virtualname__]


class ActionWrapper:
    """
    This class wraps argparse.Action instances in order to mark arguments passed
    on CLI as explicitly passed
    """

    def __init__(self, action):
        self._action = action
        functools.update_wrapper(self, action)

    def __call__(self, parser, namespace, values, option_string):
        # Let's store the call to this option as an explicit CLI call for later
        # use when overwriting any configuration settings on file with those
        # from CLI
        if getattr(parser, "_explicit_cli_args_", None) is None:
            setattr(parser, "_explicit_cli_args_", set())
        parser._explicit_cli_args_.add(
            self._action.dest
        )  # pylint: disable=protected-access
        # Carry on regular operation
        return self._action(parser, namespace, values, option_string)

    def __getattribute__(self, name):
        if name == "_action":
            return object.__getattribute__(self, name)
        # Proxy any attribute's search to the _action instance
        return getattr(self._action, name)

    def __repr__(self):
        return repr(self._action)


class ActionClassWrapper:
    """
    This class wraps argparse.Action classes in order to mark arguments passed
    on CLI as explicitly passed
    """

    def __init__(self, klass):
        self._klass = klass

    def __call__(self, *args, **kwargs):
        return ActionWrapper(self._klass(*args, **kwargs))

    def __repr__(self):
        return repr(self._klass)

    def __getattribute__(self, name):
        if name == "_klass":
            return object.__getattribute__(self, name)
        # Proxy any attributes search to the _klass instance
        return getattr(self._klass, name)


class ArgumentParser(argparse.ArgumentParser):
    def register(self, name, value, obj):  # pylint: disable=arguments-differ
        if name == "action":
            # Let's wrap it on our action class wrapper so we can latter store
            # which options were explicitly passed from CLI
            return super().register(name, value, ActionClassWrapper(obj))
        return super().register(name, value, obj)

    def parse_known_args(self, args=None, namespace=None):
        namespace, arg_strings = super().parse_known_args(args, namespace)
        explicit_cli_args = getattr(self, "_explicit_cli_args_", set())
        if "_explicit_cli_args_" not in namespace:
            setattr(namespace, "_explicit_cli_args_", set())
        namespace._explicit_cli_args_.update(explicit_cli_args)
        return namespace, arg_strings


def __init__(hub: "pop.hub.Hub"):
    """
    Set up the local memory copy of the parser
    """
    hub.conf._mem["args"] = {}


def _init_parser(hub: "pop.hub.Hub", opts):
    if "parser" not in hub.conf._mem["args"]:
        # Instantiate the parser
        hub.conf._mem["args"]["parser"] = ArgumentParser(**opts.get("_argparser_", {}))


def _keys(opts):
    """
    Return the keys in the right order
    """
    if isinstance(opts, collections.OrderedDict):
        return sorted(
            list(opts), key=lambda k: opts[k].get("display_priority", sys.maxsize)
        )
    return sorted(opts, key=lambda k: (opts[k].get("display_priority", sys.maxsize), k))


def subs(hub: "pop.hub.Hub", opts):
    """
    Set up sub parsers, if using sub parsers this needs to be called
    before calling setup.

    opts dict:
        <sub_title>:
            [desc]: 'Some subparser'
            help: 'subparser!'
    """
    _init_parser(hub, opts)
    hub.conf._mem["args"]["sub"] = hub.conf._mem["args"]["parser"].add_subparsers(
        dest="_subparser_"
    )
    hub.conf._mem["args"]["subs"] = {}
    for arg in _keys(opts):
        if arg in ("_argparser_",):
            continue
        comps = opts[arg]
        kwargs = {}
        if "help" in comps:
            kwargs["help"] = comps["help"]
        if "desc" in comps:
            kwargs["description"] = comps["desc"]
        hub.conf._mem["args"]["subs"][arg] = hub.conf._mem["args"]["sub"].add_parser(
            arg, **kwargs
        )
    return {"result": True, "return": True}


def setup(hub: "pop.hub.Hub", opts):
    """
    Take in a pre-defined opts dict and translate it to args

    opts dict:
        <arg>:
            [group]: foo
            [default]: bar
            [action]: store_true
            [options]: # arg will be turned into --arg
              - '-A'
              - '--cheese'
            [choices]:
              - foo
              - bar
              - baz
            [nargs]: +
            [type]: int
            [dest]: cheese
            help: Some great help message
    """
    _init_parser(hub, opts)
    defaults = {}
    groups = {}
    ex_groups = {}
    for arg in _keys(opts):
        if arg in ("_argparser_",):
            continue
        comps = opts[arg]
        positional = comps.pop("positional", False)
        if positional:
            args = [arg]
        else:
            long_opts = ["--{}".format(arg.replace("_", "-"))]
            short_opts = []
            for o_str in comps.get("options", []):
                if not o_str.startswith("--") and o_str.startswith("-"):
                    short_opts.append(o_str)
                    continue
                long_opts.append(o_str)
            args = short_opts + long_opts
        kwargs = {}
        kwargs["action"] = action = comps.get("action", None)

        if action is None:
            # Non existing option defaults to a StoreAction in argparse
            action = hub.conf._mem["args"]["parser"]._registry_get(
                "action", action
            )  # pylint: disable=protected-access

        if isinstance(action, str):
            signature = inspect.signature(
                hub.conf._mem["args"]["parser"]._registry_get("action", action).__init__
            )  # pylint: disable=protected-access
        else:
            signature = inspect.signature(action.__init__)

        for param in signature.parameters:
            if param == "self" or param not in comps:
                continue
            if param == "dest":
                kwargs["dest"] = comps.get("dest", arg)
                continue
            if param == "help":
                kwargs["help"] = comps.get("help", "THIS NEEDS SOME DOCUMENTATION!!")
                continue
            if param == "default":
                defaults[comps.get("dest", arg)] = comps[param]
            kwargs[param] = comps[param]

        if "group" in comps:
            group = comps["group"]
            if group not in groups:
                groups[group] = hub.conf._mem["args"]["parser"].add_argument_group(
                    group
                )
            groups[group].add_argument(*args, **kwargs)
            continue
        if "ex_group" in comps:
            group = comps["ex_group"]
            if group not in ex_groups:
                ex_groups[group] = hub.conf._mem["args"][
                    "parser"
                ].add_mutually_exclusive_group()
            ex_groups[group].add_argument(*args, **kwargs)
            continue
        if "sub" in comps:
            subs = comps["sub"]
            if not isinstance(subs, list):
                subs = [subs]
            for sub in subs:
                sparse = hub.conf._mem["args"]["subs"].get(sub)
                if not sparse:
                    # Maybe raise exception here? Malformed config?
                    continue
                sparse.add_argument(*args, **kwargs)
            continue
        hub.conf._mem["args"]["parser"].add_argument(*args, **kwargs)
    return {"result": True, "return": defaults}


def parse(
    hub: "pop.hub.Hub", args=None, namespace=None, only_parse_known_arguments=False
):
    """
    Parse the command line options
    """
    if only_parse_known_arguments:
        opts, unknown_args = hub.conf._mem["args"]["parser"].parse_known_args(
            args, namespace
        )
        opts_dict = opts.__dict__
        opts_dict["_unknown_args_"] = unknown_args
    else:
        opts = hub.conf._mem["args"]["parser"].parse_args(args, namespace)
        opts_dict = opts.__dict__
    return {"result": True, "return": opts_dict}


def render(hub: "pop.hub.Hub", defaults, cli_opts, explicit_cli_args):
    """
    For options specified as such, take the string passed into the cli and
    render it using the specified render flag
    """
    for key in explicit_cli_args:
        rend = defaults.get(key, {}).get("render")
        if rend:
            ref = f"conf.{rend}.render"
            cli_opts[key] = hub.pop.ref.last(ref)(cli_opts[key])
    return cli_opts



import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List

import pop.exc
import pop.hub
import pop.loader


def contract(
    hub: "pop.hub.Hub",  # pylint: disable=unused-argument
    raws: Iterable["pop.loader.LoadedMod"],
    mod: "pop.loader.LoadedMod",
):
    """
    Verify module level contract - functions only
    :param hub: The redistributed pop central hub
    :param raws: A list of loaded modules with contracts
    :param mod: A loader module
    """
    sig_errs = []
    sig_miss = []
    mname = mod.__name__
    for raw in raws:
        if isinstance(raw, pop.loader.LoadError):
            sig_errs.append(str(raw))
            continue
        else:
            for fun in raw._funcs:
                if fun.startswith("sig_"):
                    tfun = fun[4:]
                    if tfun not in mod._funcs:
                        sig_miss.append(tfun)
                        continue
                    sig_errs.extend(sig(mod._funcs[tfun].func, raw._funcs[fun].func))
    if sig_errs or sig_miss:
        msg = ""
        if sig_errs:
            msg += f"Signature Errors in {mname}:\n"
            for err in sig_errs:
                msg += f"{err}\n"
        if sig_miss:
            msg += f"Signature Functions Missing in {mname}:\n"
            for err in sig_miss:
                msg += f"{err}\n"
        msg = msg.strip()
        raise pop.exc.ContractSigException(msg)


def sig_map(ver: Callable) -> Dict[str, Any]:
    """
    Generates the map dict for the signature verification
    """
    vsig = inspect.signature(ver)
    vparams = list(vsig.parameters.values())
    vdat = {"args": [], "v_pos": -1, "kw": [], "kwargs": False, "ann": {}}
    for ind in range(len(vparams)):
        param = vparams[ind]
        val = param.kind.value
        name = param.name
        if val == 0 or val == 1:
            vdat["args"].append(name)
            if param.default != inspect._empty:  # Is a KW, can be inside of **kwargs
                vdat["kw"].append(name)
        elif val == 2:
            vdat["v_pos"] = ind
        elif val == 3:
            vdat["kw"].append(name)
        elif val == 4:
            vdat["kwargs"] = ind
        if param.annotation != inspect._empty:
            vdat["ann"][name] = param.annotation
    return vdat


def sig(func: Callable, ver: Callable) -> List[str]:
    """
    Takes 2 functions, the first function is verified to have a parameter signature
    compatible with the second function
    """
    errors = []
    fsig = inspect.signature(func)
    fparams = list(fsig.parameters.values())
    vdat = sig_map(ver)
    arg_len = len(vdat["args"])
    v_pos = False
    for ind in range(len(fparams)):
        param = fparams[ind]
        val = param.kind.value
        name = param.name
        has_default = param.default != inspect._empty
        ann = param.annotation
        vann = vdat["ann"].get(name, inspect._empty)
        if vann != ann:
            errors.append(f'Parameter, "{name}" is type "{str(ann)}" not "{str(vann)}"')
        if val == 2:
            v_pos = True
        if val == 0 or val == 1:
            if ind >= arg_len:  # Past available positional args
                if not vdat["v_pos"] == -1:  # Has a *args
                    if ind >= vdat["v_pos"] and v_pos:
                        # Invalid unless it is a kw
                        if not name in vdat["kw"]:
                            # Is a kw
                            errors.append(f'Parameter "{name}" is invalid')
                        if vdat["kwargs"] is False:
                            errors.append(f'Parameter "{name}" not defined as kw only')
                        continue
                elif vdat["kwargs"] is not False and not has_default:
                    errors.append(
                        f'Parameter "{name}" is past available positional params'
                    )
                elif vdat["kwargs"] is False:
                    errors.append(
                        f'Parameter "{name}" is past available positional params'
                    )
            else:
                v_param = vdat["args"][ind]
                if v_param != name:
                    errors.append(
                        f'Parameter "{name}" does not have the correct name: {v_param}'
                    )
        if val == 2:
            if ind < vdat["v_pos"]:
                errors.append(
                    f'Parameter "{name}" is not in the correct position for *args'
                )
        if val == 3:
            if name not in vdat["kw"] and not vdat["kwargs"]:
                errors.append(f'Parameter "{name}" is not available as a kwarg')
        if val == 4:
            if vdat["kwargs"] is False:
                errors.append(f"Kwargs are not permitted as a parameter")
    return errors



"""
The Proc sub is used to spin up worker processes that run hub referenced
coroutines.
"""
import asyncio
import atexit
import itertools
import os
import subprocess
import sys

import msgpack

import pop.hub


def __init__(hub: "pop.hub.Hub"):
    """
    Create constants used by the client and server side of procs
    """
    hub.proc.DELIM = b"d\xff\xcfCO)\xfe="
    hub.proc.D_FLAG = b"D"
    hub.proc.I_FLAG = b"I"
    hub.proc.Workers = {}
    hub.proc.WorkersIter = {}
    hub.proc.WorkersTrack = {}


def _get_cmd(hub: "pop.hub.Hub", ind, ref, ret_ref, sock_dir):
    """
    Return the shell command to execute that will start up the worker
    """
    code = "import sys; "
    code += "import pop.hub; "
    code += "hub = pop.hub.Hub(); "
    code += 'hub.pop.sub.add("pop.mods.proc"); '
    code += f'hub.proc.worker.start("{sock_dir}", "{ind}", "{ref}", "{ret_ref}")'
    cmd = f"{sys.executable} -c '{code}'"
    return cmd


def mk_proc(hub: "pop.hub.Hub", ind, workers, ret_ref, sock_dir):
    """
    Create the process and add it to the passed in workers dict at the
    specified index
    """
    ref = os.urandom(3).hex() + ".sock"
    workers[ind] = {"ref": ref}
    workers[ind]["path"] = os.path.join(sock_dir, ref)
    cmd = _get_cmd(hub, ind, ref, ret_ref, sock_dir)
    workers[ind]["proc"] = subprocess.Popen(cmd, shell=True)
    workers[ind]["pid"] = workers[ind]["proc"].pid


async def pool(
    hub: "pop.hub.Hub", num, name: str = "Workers", callback=None, sock_dir=None
):
    """
    Create a new local pool of process based workers

    :param num: The number of processes to add to this pool
    :param ref: The location on the hub to create the Workers dict used to
        store the worker pool, defaults to `hub.pop.proc.Workers`
    :param callback: The pop ref to call when the process communicates
        back
    """
    ret_ref = os.urandom(3).hex() + ".sock"
    ret_sock_path = os.path.join(sock_dir, ret_ref)
    if not hasattr(hub.proc, "Tracker"):
        hub.proc.init.mk_tracker()
    workers = {}
    if callback:
        await asyncio.start_unix_server(
            hub.proc.init.ret_work(callback), path=ret_sock_path
        )
    for ind in range(num):
        hub.proc.init.mk_proc(ind, workers, ret_ref, sock_dir)
    w_iter = itertools.cycle(workers)
    hub.proc.Workers[name] = workers
    hub.proc.WorkersIter[name] = w_iter
    hub.proc.WorkersTrack[name] = {"subs": [], "ret_ref": ret_ref, "sock_dir": sock_dir}
    up = set()
    while True:
        for ind in workers:
            if os.path.exists(workers[ind]["path"]):
                up.add(ind)
        if len(up) == num:
            break
        await asyncio.sleep(0.01)
    # TODO: This seems to be spawning extra procs, this should be fixed
    # asyncio.ensure_future(hub.proc.init.maintain(name))


async def maintain(hub: "pop.hub.Hub", name):
    """
    Keep an eye on these processes
    """
    workers = hub.proc.Workers[name]
    while True:
        for ind, data in workers.items():
            if not data["proc"].poll():
                hub.proc.init.mk_proc(ind, workers)
        await asyncio.sleep(2)


def mk_tracker(hub: "pop.hub.Hub"):
    """
    Create the process tracker, this simply makes a data structure to hold
    process references and sets them to be terminated when the system is
    shutdown.
    """
    hub.proc.Tracker = True
    atexit.register(hub.proc.init.clean)


def clean(hub: "pop.hub.Hub"):
    """
    Clean up the processes registered in the tracker
    """
    for name, workers in hub.proc.Workers.items():
        for ind in workers:
            workers[ind]["proc"].terminate()


def ret_work(hub: "pop.hub.Hub", callback):
    async def work(reader, writer):
        """
        Process the incoming work
        """
        inbound = await reader.readuntil(hub.proc.DELIM)
        inbound = inbound[: -len(hub.proc.DELIM)]
        payload = msgpack.loads(inbound, raw=False)
        ret = await callback(payload)
        ret = msgpack.dumps(ret, use_bin_type=True)
        ret += hub.proc.DELIM
        writer.write(ret)
        await writer.drain()
        writer.close()

    return work



"""
Define the JSON loader interface
"""
import json

import pop.hub

__virtualname__ = "json"
__contracts__ = [__virtualname__]


def __virtual__(hub):
    return True


def load(hub: "pop.hub.Hub", path):
    """
    Use json to read in a file
    """
    try:
        with open(path) as fp_:
            ret = json.loads(fp_.read())
        return ret
    except FileNotFoundError:
        return {}


def render(hub: "pop.hub.Hub", val):
    """
    Take the string and render it in json
    """
    return json.loads(val)



"""
This sub is used to set up logging for pop projects and injects logging
options into conf making it easy to add robust logging
"""
import logging
from typing import Any
from typing import Dict

import pop.hub


def __init__(hub: "pop.hub.Hub"):
    """
    Set up variables used by the log subsystem
    """
    hub.conf.log.LEVELS = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }


def conf(hub: "pop.hub.Hub", name: str) -> Dict[str, Any]:
    """
    Return the conf dict for logging, this should be merged OVER by the loaded
    config dict(s)
    """
    # TODO: Make this more robust to handle more logging interfaces
    ldict = {
        "log_file": {
            "default": f"{name}.log",
            "help": "The location of the log file",
            "group": "Logging Options",
        },
        "log_level": {
            "default": "warning",
            "help": "Set the log level, either quiet, info, warning, or error",
            "group": "Logging Options",
        },
        "log_fmt_logfile": {
            "default": "%(asctime)s,%(msecs)03d [%(name)-17s][%(levelname)-8s] %(message)s",
            "help": "The format to be given to log file messages",
            "group": "Logging Options",
        },
        "log_fmt_console": {
            "default": "[%(levelname)-8s] %(message)s",
            "help": "The log formatting used in the console",
            "group": "Logging Options",
        },
        "log_datefmt": {
            "default": "%H:%M:%S",
            "help": "The date format to display in the logs",
            "group": "Logging Options",
        },
        "log_plugin": {
            "default": "basic",
            "help": "The logging plugin to use",
            "group": "Logging Options",
        },
    }
    return ldict
